
#################################################
#################################################
##          COPYRIGHT - DATA SOCIETY           ##
#################################################
#################################################

#############################
##      SQL + PYTHON       ##
#############################

## NOTE: To run code blocks (separated by #%%), select somewhere inside the block 
##       and press ctrl + enter for Windows or Linux, 
##       and command + enter for Macs

#==================================================================================
"""
The pursose of this script is to create and populate a database on your
local MySQL server. All but one of the tables are populated with data from a csv
file. The family_member_interest table is created with pure python code, to randomly
generate interests for different people.

Note: You can run this script as many times as you like, because each time it
drops and recreates all of the tables. This style of loading process is called
a 'destructive load,' as opposed to an 'incremental load', which does not
drop everything first.
"""

#%%
#### Slide 93: Connecting to a database ####

# Import the MySQL connector library
import mysql.connector

# Open a connection and save it to a variable
dbconn = mysql.connector.connect(user='ta_anna',        #<- type in your username
                                  password='',             #<- leave password blank if 
                                                           #<- you didn't create one
                                  host='127.0.0.1',        #<- host IP address
                                  database='familydb')     #<- name of the database we created

"""
Notice that the dbconn variable has type MySQLConnection. 
This object holds the connection to the database. To run 
commands or SELECT statements, we need to create cursor object
from this connection:
"""
cur = dbconn.cursor()

#==================================================================================
#%%
#### Slides 96 - 99: Creating tables in database ####

"""
We want to crete several tables using this python script. 
There are many ways to pass the SQL create statements to 
the database. Ultimately each statement must be executed 
with cur.execute(). The below dictionary stores the SQL
to create each fo the tables, which will allow us to easily
loop over them. 
"""

tables=[
{
'name': 'family',
'query': """
create table family(
    id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,
    last_name VARCHAR(50) NOT NULL
)
"""
},
{
'name': 'interest',
'query': """
create table interest(
    id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,
    interest VARCHAR(50) NOT NULL
)
"""
},
{
 'name': 'family_member',
 'query': """
 create table family_member(
    id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,
    first_name varchar(50) NOT NULL,
    relation CHAR(1) NOT NULL,
    birth_date DATE NOT NULL,
    current_age INT NOT NULL,
    family_id INT NOT NULL
    )
"""
 },
{
 'name': 'family_member_interest',
 'query': """
 CREATE TABLE family_member_interest(
    id INT PRIMARY KEY NOT NULL AUTO_INCREMENT,
    family_member_id INT NOT NULL,
    interest_id INT NOT NULL
    )
 """ 
 }
]

#==================================================================================
#%%
#### Slides 100 - 102: Defining foreign keys ####

"""
The main feature of relational databases is to store records that are related in some way. 
 In our example we have a `family` table and a `family_member` table. Each family_member 
 belongs to a family. We could specify each family_member's last name in the family_member 
 table, but it will save us storage space in the database to simply store a reference to 
 the family table. We put the query statement to create each of the foreign keys in the below
 query, so that we can loop through them later.


The final keyword `ON DELETE` instructs MySQL what to do if someone's family is removed
 from the `family` table. Setting this option to `CASCADE` means that if a family is 
 removed from the `family` table, we want to "cascade" this removal to all children. 
 In other words, remove all corresponding family members from the `family_member` table. 


The table 'family_member_interest' has two 
 foreign keys: one foreign key connects family_member and
 the other connects to interest.
"""
foreign_keys = [
      
  {
   'tablename':'family_member',
   'fks':[
           {
            'name': 'fk_family',
            'query': """
              ALTER TABLE family_member
                ADD CONSTRAINT fk_family
                FOREIGN KEY (family_id)
                    REFERENCES family(id)
                    ON DELETE CASCADE
              """
          }
  ]}
 ,
 {
  'tablename': 'family_member_interest',
  'fks':[
          {
           'name': 'fk_pi_family_member',
           'query':"""
              ALTER TABLE family_member_interest 
                 ADD CONSTRAINT fk_pi_family_member
                FOREIGN KEY (family_member_id)
                    REFERENCES family_member(id)
                    ON DELETE CASCADE
            """
            },
          {
          'name': 'fk_pi_interest',
          'query': """
               ALTER TABLE family_member_interest
               ADD CONSTRAINT fk_pi_interest
                FOREIGN KEY (interest_id)
                    REFERENCES interest(id)
                    ON DELETE CASCADE
              """
          }
           
  ]}
 ]

#==================================================================================
#%%
#### Slide 103: Refresh the database ####

"""
To refresh the database, we'll need to drop all of the tables and 
then recreate them, and then populate them with data. Before we can
drop the tables, we should remove any foreign key constraints. We 
do this to avoid errors encountered by trying to drop a parent table
before its children. The cleanest way to avoid this error is to 
simply drop the relationship, and then the tables can be dropped in 
any order.
"""

"""
The first time you run this script, no tables (and therefore no foreign
 keys) will exist yet. So we'll make some helper functions to tell
 us whether or not the foreign key exists, so that we don't cause
 and error by attempting to delete a foreign key that doesn't exist.

To check if a foreign key exists, we can check against the 
table_constraints table of information_schema. 
"""
def fk_exists(tbl_name, fk_name):
    cur.execute("""select * from information_schema.table_constraints
                where table_schema='familydb'
                and table_name='{}'
                and constraint_name = '{}'
                """.format(tbl_name, fk_name))
    # if the foreign key exists, this query will return it.
    # otherwise, there will be 0 results
    return len(cur.fetchall()) > 0

"""
MySQL automatically creates indexes on any foreign key column. So
we'll also check if an index with the same name exists. This helper
function does that for us, making use of the statistics table
in information_schema, which contains all table indexes.
"""
def index_exists(tbl_name, index_name):
    cur.execute("""select * from information_schema.statistics
                where table_schema='familydb'
                and table_name='{}'
                and index_name = '{}'
                """.format(tbl_name, index_name))
    # if the index exists, this query will return it.
    # otherwise, there will be 0 results
    return len(cur.fetchall()) > 0


#==================================================================================
#%%
#### Slide 104: Dropping foreign key constraints ####

"""
Now we can loop through the foreign_keys list. For each table, 
thanks to the way we designed this list, we can easily get
the name of the foreign keys in order to drop them.
Before executing the command to drop the key and index, 
we first use the helper function to make sure they indeed
exist, as trying to drop something that doesn't exist will 
throw an error.
"""
for tbl in foreign_keys:
    tbl_name = tbl['tablename']
    fks = tbl['fks']
    
    for fk in fks:
        drop_fk = 'ALTER TABLE {} DROP FOREIGN KEY {}'
        drop_index = 'DROP INDEX {} ON {}' # mysql automatically indexes fk columns
        key_name = fk['name']
        
        if fk_exists(tbl_name, key_name):
            cur.execute(drop_fk.format(tbl_name, key_name))
            
        if index_exists(tbl_name, key_name):
            cur.execute(drop_index.format(key_name, tbl_name))
        
dbconn.commit()


#==================================================================================
#%%
#### Slide 105: Recreating the tables ####

"""
Now that we've dropped all of the foreign keys, we can drop the
tables one by one
"""
for table in tables:
    print('dropping '+table['name'])
    cur.execute('drop table if exists '+table['name'])
dbconn.commit()
    
#%%
"""
Create each table one by one using the stored CREATE TABLE
statements
"""
for table in tables:
    print('creating '+table['name'])
    cur.execute(table['query'])
dbconn.commit()


#%%
"""
Now that the tables are created, we can add the foreign
keys one by one.
"""
for tbl in foreign_keys:
    tbl_name = tbl['tablename']
    fks = tbl['fks']
    for fk in fks:
        cur.execute(fk['query'])

dbconn.commit()

#==================================================================================
#%%
#### Slides 106 - 107: Loading files into a database ####

"""
Now we load the tables with some locally stored data. Here we
use the base python file reader, but for more sophisticated
file handling, you can use the various file input functions 
in the pandas library
"""
def load_file(tablename):
    f = open('data/'+tablename+'.csv', 'r')
    header = f.readline()
    
    query = """
    insert into {}({})
     values({})
    """
    
    line = f.readline().strip()
    n = 0
    while line:
        n += 1
        line_parsed = ','.join(["'"+val+"'" for val in line.split(',')])
        cur.execute(query.format(tablename, header, line_parsed))
        line = f.readline().strip()
  
    
    # print a helpful message when file loaded
    print('loaded {} lines to {}'.format(n, tablename))
    
    """
    we return some information about the file: namely the number
    of records we just loaded.
    """
    return n


n_families = load_file('family')
n_fam_members = load_file('family_member')
n_interests = load_file('interest')
dbconn.commit() # only commit once all 3 loaded successfully

#==================================================================================
#%%
#### Slide 108: Updating ages ####

"""
update the current_age field to calculate their age as of today.
we'll need the DATEDIFF function, and the MySQL internal CURRENT_DATE variable
"""
cur.execute("update family_member set current_age = floor( DATEDIFF(CURRENT_DATE, birth_date)/365) ")
dbconn.commit()


#==================================================================================
#%%
#### Slide 109: Creating people's interests ####

# make some random interests for people
import numpy as np
np.random.seed(100)

n_fam_mem_interests = 0
for i in range(n_fam_members):
  n_interests_i = np.random.randint(0,6,1)[0] # create somewhere between 0 and 5 interests per person
  interests_i = np.random.randint(1, n_interests, n_interests_i)
  for j in interests_i:
    query = """
    insert into family_member_interest(family_member_id, interest_id)
    values({},'{}')
    """.format(i, j)
    cur.execute(query)
    n_fam_mem_interests += 1

dbconn.commit()

print('loaded {} lines to family_member_interest'.format(n_fam_mem_interests))
